package App::ABAN::DumbUI;
#ABSTRACT: ABAN text-mode UI for a "dumb" terminal
use 5.028;
use warnings qw(all);
use strict;

use POE qw(Wheel::ReadWrite Filter::Stream);
use Term::ReadKey qw(ReadMode);
use Data::Dump qw(pp);
use Text::ASCIITable;


sub create {
	POE::Session->create(
		package_states  => [
			(__PACKAGE__) => {
				_start => '_start',
				_stop => '_stop',
				keystroke => '_keystroke',
				
				# our API
				set_status => '_set_status',
				set_wipeinfo => '_set_wipeinfo',
				set_protected => '_set_protected',
				set_drives => '_set_drives',
				set_victims => '_set_victims',
				set_benchmark => '_set_benchmark',
				set_prf => '_set_prf',

				wipe_progress => '_wipe_progress',
				wipe_complete => '_wipe_complete',

				shutdown => '_shutdown',

			},
		]
	);
}

sub _start {
	$_[KERNEL]->alias_set('UI');
	$_[HEAP]{StdIO} = POE::Wheel::ReadWrite->new(
		InputHandle => \*STDIN,
		OutputHandle => \*STDOUT,
		InputFilter => POE::Filter::Stream->new(),
		OutputFilter => POE::Filter::Line->new(Literal => "\n"),
		InputEvent => 'keystroke',
	);
	ReadMode 'cbreak';

	$_[KERNEL]->post(MAIN => 'ui_ready');
}

sub _stop {
	ReadMode 'restore';
}

sub _keystroke {
	my $key = $_[ARG0];
	my $io = $_[HEAP]{StdIO};

	if ('h' eq $key || '?' eq $key) {
		$io->put(<<KEY_HELP);
Keys:
   h, ? - this message
   s - re-display last status message
   v - display list of devices we're wiping (victims)
   p - display list of protected devices (what we're booted from)
   d - display list of possible devices (wipe unless protected)
   b - display PRF benchmark results

Control-c: Abort. Note that data already wiped is unrecoverable. If
           aborted after the wipe has started, just results in a partial job.
KEY_HELP
	} elsif ('s' eq $key) {
		$io->put("Current status: $_[HEAP]{app_status}");
	} elsif ('p' eq $key) {
		$io->put("Protected devices: " . pp($_[HEAP]{app_protected}));
	} elsif ('d' eq $key) {
		$io->put("All potential disks: " . pp($_[HEAP]{app_drives}));
	} elsif ('v' eq $key) {
		$io->put("Victims: " . pp($_[HEAP]{app_victims}));
	} elsif ('b' eq $key) {
		$io->put("PRF Benchmark: " . pp($_[HEAP]{app_benchmark}));
		$io->put("PRF Chosen: $_[HEAP]{app_prf}.");
	} else {
		$io->put("Unknown key: $key");
	}

	return;
}

sub _shutdown {
	$_[HEAP]{StdIO}->flush; # give it a try...
	$_[HEAP]{StdIO} = undef; # session idle after this, will exit.
}

sub _set_status {
	$_[HEAP]{app_status} = $_[ARG0];
	$_[HEAP]{StdIO}->put("Status: $_[ARG0]");
	return;
}

sub _set_wipeinfo {
	$_[HEAP]{app_wipeinfo} = $_[ARG0];
	return;
}

sub _set_protected {
	$_[HEAP]{app_protected} = $_[ARG0];
	$_[HEAP]{StdIO}->put("Received protected device list; p to display");
	return;
}

sub _set_drives {
	$_[HEAP]{app_drives} = $_[ARG0];
	$_[HEAP]{StdIO}->put("Received drive list; d to display");
	return;
}

sub _set_victims {
	$_[HEAP]{app_victims} = $_[ARG0];
	$_[HEAP]{StdIO}->put("Received list of drives to wipe (victims); v to re-display:", pp $_[ARG0]);
	return;
}

sub _set_benchmark {
	$_[HEAP]{app_benchmark} = $_[ARG0];
	$_[HEAP]{StdIO}->put("Received PRF benchmark; b to display");
	return;
}

sub _set_prf {
	$_[HEAP]{app_prf} = $_[ARG0];
	$_[HEAP]{StdIO}->put("Chosen PRF is $_[ARG0].");
	return;
}

sub _wipe_progress {
	my ($ident, $msg) = @_[ARG0..ARG1];

	$_[HEAP]{StdIO}->put("<$ident> $msg");
}

sub _wipe_complete {
	$_[HEAP]{StdIO}->put("ABAN is done wiping disks.");

	my $t = Text::ASCIITable->new(
		{headingText => "Final ABAN Disk Wiper Results"});
	$t->setCols("Disk", "Speed(W)", "Speed(V)", "Unwiped", "Unverified",
		"Notes");
	$t->alignCol("Unwiped", 'right');
	$t->alignCol("Unverified", 'right');
	
	foreach my $info (sort { $a->{victim}{node} cmp $b->{victim}{node} }
		@{$_[HEAP]{app_wipeinfo}})
	{
		$t->addRow(
			$info->{victim}{node},
			$info->{wipe_complete_info}{rate},
			$info->{verify_complete_info}{rate},
			$info->{wipe_complete_info}{worklist},
			$info->{verify_complete_info}{worklist},
			($info->{errored_info} ? "DIED ($info->{errored_info})" : ""),
		);
	}

	$_[HEAP]{StdIO}->put("$t");

	return;
}

1;

__END__

=pod

=encoding UTF-8

=head1 NAME

App::ABAN::DumbUI - ABAN text-mode UI for a "dumb" terminal

=head1 VERSION

version 0.020

=head1 AUTHOR

Anthony DeRobertis <anthony@derobert.net>

=head1 COPYRIGHT AND LICENSE

Copyright (C)2019-2020 Customer Relationship Metrics.
Copyright (C)2019-2022 Anthony DeRobertis.
Web page modern style (C)2020 Ruben Albornoz.

This program is free software; you can redistribute it and/or modify it
under the terms of the Perl Artistic License or the GNU General Public
License as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

In addition, if you are distributing under the GPL, as a special
exception, the above copyright holders give permission to link the code
of this program with the OpenSSL Library (or with modified versions of
OpenSSL that use the same license as OpenSSL), and distribute linked
combinations including the two. You must obey the GNU General Public
License in all respects for all of the code used other than OpenSSL.

This exception applies to versions of OpenSSL under the traditional
(pre-3.0) OpenSSL licensing, versions under Apache License v2, and any
future license OpenSSL may be released under.

If you modify this program, you may extend this exception to your
version, but you are not obligated to do so. If you do not wish to do
so, delete this exception statement from your version.

=cut
